from datetime import datetime
from typing import List, Optional

from sqlalchemy import (
    CHAR,
    Column,
    ForeignKey,
    MetaData,
    String,
    Table,
    func,
)
from sqlalchemy.dialects.oracle import BINARY_FLOAT, NUMBER
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    column_property,
    mapped_column,
    relationship,
)

from .enum_type import EnumType
from .enums import CodeEnum
from .fields import UTCDateTime

OLDW_METADATA = MetaData(schema="OLDW")

course_subject = Table(
    "CRSE_SBJCT",
    OLDW_METADATA,
    Column("CRSE_KEY", ForeignKey("CRSE.CRSE_KEY"), primary_key=True),
    Column("SBJCT_KEY", ForeignKey("SBJCT.SBJCT_KEY"), primary_key=True),
)

course_bundle = Table(
    "CRSE_BUNDLE",
    OLDW_METADATA,
    Column("CRSE_KEY", ForeignKey("CRSE.CRSE_KEY"), primary_key=True),
    Column("BUNDLE_KEY", ForeignKey("BUNDLE.BUNDLE_KEY"), primary_key=True),
)


bundle_subject = Table(
    "BUNDLE_SBJCT",
    OLDW_METADATA,
    Column("BUNDLE_KEY", ForeignKey("BUNDLE.BUNDLE_KEY"), primary_key=True),
    Column("SBJCT_KEY", ForeignKey("SBJCT.SBJCT_KEY"), primary_key=True),
)

assessment_item = Table(
    "ASSESSMENT_ITEM",
    OLDW_METADATA,
    Column("ASSESSMENT_KEY", ForeignKey("ASSESSMENT.ASSESSMENT_KEY"), primary_key=True),
    Column("ITEM_KEY", ForeignKey("ITEM.ITEM_KEY"), primary_key=True),
)


def inst_repr(typ, attrs):
    params = [f"{key}='{value}'" for key, value in attrs.items() if value is not None]
    return f"<{typ}({', '.join(params)})>"


class Base(DeclarativeBase):
    metadata = OLDW_METADATA
    __repr_fields__ = ["id"]

    def __repr__(self):
        return inst_repr(
            self.__class__.__name__,
            {key: getattr(self, key) for key in self.__repr_fields__},
        )


class SourceSystem(Base):
    __tablename__ = "SRC_SYS"
    __repr_fields__ = ["id", "platform_name", "description"]

    id: Mapped[str] = mapped_column("SRC_SYS_CD", String(50), primary_key=True)
    description: Mapped[str] = mapped_column("SRC_SYS_DES", String(100), nullable=False)
    platform_name: Mapped[str] = mapped_column(
        "LRNG_PLTFRM_NM", String(100), nullable=False
    )
    load_date: Mapped[Optional[datetime]] = mapped_column(
        "LOAD_DT", UTCDateTime, nullable=True
    )

    courses: Mapped[List["Course"]] = relationship(back_populates="source_system")
    users: Mapped[List["OnlineLearningUser"]] = relationship(
        back_populates="source_system"
    )
    subjects: Mapped[List["Subject"]] = relationship(back_populates="source_system")
    bundles: Mapped[List["Bundle"]] = relationship(back_populates="source_system")
    assessments: Mapped[List["Assessment"]] = relationship(
        back_populates="source_system"
    )
    questions: Mapped[List["Question"]] = relationship(back_populates="source_system")


class Course(Base):
    __tablename__ = "CRSE"
    __repr_fields__ = ["id", "name", "slug", "source_system_id"]

    class LevelCode(CodeEnum):
        BEGINNER = "BEGINNER", "Beginner"
        INTERMEDIATE = "INTERMEDIATE", "Intermediate"
        ADVANCED = "ADVANCED", "Advanced"

    id: Mapped[str] = mapped_column("CRSE_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_course_id: Mapped[str] = mapped_column("SRC_SYS_CRSE_CD", String(255))
    name: Mapped[str] = mapped_column("CRSE_NM", String(2000))
    slug: Mapped[Optional[str]] = mapped_column(
        "CRSE_SLUG_TXT", String(4000), nullable=True
    )
    launched: Mapped[Optional[datetime]] = mapped_column(
        "CRSE_LAUNCH_DTTM", UTCDateTime, nullable=True
    )
    image_url: Mapped[Optional[str]] = mapped_column(
        "CRSE_IMAGE_URL_TXT", String(4000), nullable=True
    )
    video_url: Mapped[Optional[str]] = mapped_column(
        "CRSE_VIDEO_URL_TXT", String(4000), nullable=True
    )
    description: Mapped[Optional[str]] = mapped_column(
        "CRSE_DES", String(4000), nullable=True
    )
    html_description: Mapped[Optional[str]] = mapped_column(
        "CRSE_HTML_DES", String(4000), nullable=True
    )
    level_code: Mapped[Optional[str]] = mapped_column(
        "CRSE_LVL_CD", EnumType(LevelCode, String(50)), nullable=True
    )
    res_ind: Mapped[Optional[str]] = mapped_column(
        "CRSE_RES_IND", String(1), nullable=True
    )

    source_system: Mapped[SourceSystem] = relationship(back_populates="courses")
    classes: Mapped[List["Class"]] = relationship(back_populates="course")
    ratings: Mapped[List["CourseRating"]] = relationship(back_populates="course")
    course_instances: Mapped[List["CourseInstance"]] = relationship(
        back_populates="course"
    )
    subjects: Mapped[List["Subject"]] = relationship(
        secondary=course_subject, back_populates="courses"
    )
    bundles: Mapped[List["Bundle"]] = relationship(
        secondary=course_bundle, back_populates="courses"
    )
    course_memberships: Mapped[List["CourseMembership"]] = relationship(
        back_populates="course"
    )


class Class(Base):
    __tablename__ = "CLASS"
    __repr_fields__ = ["id", "course_id", "source_system_id"]

    id: Mapped[str] = mapped_column("CLASS_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_class_id: Mapped[str] = mapped_column("SRC_SYS_CLASS_CD", String(255))
    course_id: Mapped[str] = mapped_column(
        "CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY")
    )
    begin_date: Mapped[datetime] = mapped_column("CLASS_BEGIN_DTTM", UTCDateTime)
    end_date: Mapped[Optional[datetime]] = mapped_column(
        "CLASS_END_DTTM", UTCDateTime, nullable=True
    )
    enrollment_begin_date: Mapped[Optional[datetime]] = mapped_column(
        "CLASS_ENRL_PRD_BEGIN_DTTM", UTCDateTime, nullable=True
    )
    enrollment_end_date: Mapped[Optional[datetime]] = mapped_column(
        "CLASS_ENRL_PRD_END_DTTM", UTCDateTime, nullable=True
    )
    course_instance_id: Mapped[str] = mapped_column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    is_dummy_class: Mapped[Optional[int]] = mapped_column(
        "DUMMY_CLASS_IND", NUMBER(1), nullable=True
    )

    course_instance: Mapped["CourseInstance"] = relationship(back_populates="classes")
    course: Mapped["Course"] = relationship(back_populates="classes")
    class_memberships: Mapped[List["ClassMembership"]] = relationship(
        back_populates="class_"
    )
    modules: Mapped[List["Module"]] = relationship(back_populates="class_")
    submodules: Mapped[List["Submodule"]] = relationship(back_populates="class_")
    items: Mapped[List["Item"]] = relationship(back_populates="class_")


class OnlineLearningUser(Base):
    __tablename__ = "ONLINE_LRNG_USR"
    __repr_fields__ = ["id", "full_name", "source_system_id", "source_system_user_id"]

    id: Mapped[str] = mapped_column("USR_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_user_id: Mapped[str] = mapped_column("SRC_SYS_USR_CD", String(50))
    last_login: Mapped[Optional[datetime]] = mapped_column(
        "SRC_SYS_LAST_LOGIN_DTTM", UTCDateTime, nullable=True
    )
    date_joined: Mapped[Optional[datetime]] = mapped_column(
        "LRNG_PLTFRM_JOIN_DTTM", UTCDateTime, nullable=True
    )
    first_name: Mapped[Optional[str]] = mapped_column(
        "FIRST_NM", String(100), nullable=True
    )
    last_name: Mapped[Optional[str]] = mapped_column(
        "LAST_NM", String(100), nullable=True
    )
    full_name: Mapped[Optional[str]] = mapped_column(
        "FULL_NM", String(200), nullable=True
    )

    source_system: Mapped[SourceSystem] = relationship(back_populates="users")
    class_memberships: Mapped[List["ClassMembership"]] = relationship(
        back_populates="user"
    )
    email_addresses: Mapped[List["EmailAddress"]] = relationship(back_populates="user")
    course_ratings: Mapped[List["CourseRating"]] = relationship(back_populates="user")
    bundle_memberships: Mapped[List["BundleMembership"]] = relationship(
        back_populates="user"
    )
    course_memberships: Mapped[List["CourseMembership"]] = relationship(
        back_populates="user"
    )


class RoleCode(CodeEnum):
    LEARNER = "LEARNER", "Learner"
    RESEARCHER = "RESEARCHER", "Researcher"
    ADMINISTRATOR = "ADMINISTRATOR", "Administrator"
    BROWSER = "BROWSER", "Browser"
    TEACHING_STAFF = "TEACHING_STAFF", "Teaching Staff"
    COURSE_ASSISTANT = "COURSE_ASSISTANT", "Course Assistant"
    UNENROLLED_LEARNER = "UNENROLLED_LEARNER", "Unenrolled Learner"
    PREENROLLED_LEARNER = "PRE_ENROLLED_LEARNER", "Pre-enrolled Learner"
    INSTRUCTOR = "INSTRUCTOR", "Instructor"
    MENTOR = "MENTOR", "Mentor"


class CertificateType(CodeEnum):
    VERIFIED = "VERIFIED", "Verified"
    AUDIT = "AUDIT", "Audit"
    HONORS = "HONOR", "Honors"


class PaymentSourceCode(CodeEnum):
    FINANCIAL_AID_AWARD = "FINANCIAL AID AWARD", "Financial Aid Award"
    ENTERPRISE = "ENTERPRISE", "Enterprise"
    INDIVIDUAL_PAYMENT = "INDIVIDUAL PAYMENT", "Individual Payment"


class ClassMembership(Base):
    __tablename__ = "CLASS_MBRSHP"
    __repr_fields__ = ["id", "role_code", "class_id", "user_id"]

    id: Mapped[str] = mapped_column("CLASS_MBRSHP_KEY", CHAR(36), primary_key=True)
    class_id: Mapped[str] = mapped_column(
        "CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY")
    )
    user_id: Mapped[str] = mapped_column(
        "USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY")
    )
    role_code: Mapped[str] = mapped_column(
        "MBRSHP_ROLE_CD", EnumType(RoleCode, String(50))
    )
    begin_date: Mapped[datetime] = mapped_column("MBRSHP_BEGIN_DTTM", UTCDateTime)
    eligible_certificate_type: Mapped[Optional[str]] = mapped_column(
        "ELIG_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    awarded_certificate_type: Mapped[Optional[str]] = mapped_column(
        "AWARD_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    cert_award_date: Mapped[Optional[datetime]] = mapped_column(
        "CERT_AWARD_DTTM", UTCDateTime, nullable=True
    )
    payment_source_code: Mapped[Optional[str]] = mapped_column(
        "PMT_SRC_CD", EnumType(PaymentSourceCode, String(50)), nullable=True
    )

    class_: Mapped[Class] = relationship(back_populates="class_memberships")
    user: Mapped[OnlineLearningUser] = relationship(back_populates="class_memberships")


class EmailAddress(Base):
    __tablename__ = "USR_EMAIL_ADDR"
    __repr_fields__ = ["user_id", "email", "primary_address"]

    user_id: Mapped[str] = mapped_column(
        "USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY"), primary_key=True
    )
    email: Mapped[str] = mapped_column(
        "USR_EMAIL_ADDR_TXT", String(500), primary_key=True
    )
    primary_address: Mapped[int] = mapped_column("PRMRY_EMAIL_ADDR_IND", NUMBER(1))

    user: Mapped[OnlineLearningUser] = relationship(back_populates="email_addresses")

    username: Mapped[str] = column_property(
        func.substr(email, 0, func.instr(email, "@", 1) - 1), deferred=True
    )
    domain: Mapped[str] = column_property(
        func.substr(email, func.instr(email, "@", 1) + 1), deferred=True
    )


class CourseRating(Base):
    __tablename__ = "USR_CRSE_RT"
    __repr_fields__ = ["id", "user_id", "course_id", "rating", "max_rating"]

    class RatingSystemCode(CodeEnum):
        STAR = "STAR", "Star"
        NPS_FIRST_WEEK = "NPS_FIRST_WEEK", "Net Promoter Score, First Week"
        NPS_END_OF_COURSE = "NPS_END_OF_COURSE", "Net Promoter Score, End of Course"

    id: Mapped[str] = mapped_column("USR_CRSE_RT_KEY", CHAR(36), primary_key=True)
    user_id: Mapped[str] = mapped_column(
        "USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY")
    )
    course_id: Mapped[str] = mapped_column(
        "CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY")
    )
    date: Mapped[datetime] = mapped_column("CRSE_RT_DTTM", UTCDateTime)
    rating_code: Mapped[Optional[str]] = mapped_column(
        "RT_SYS_CD", EnumType(RatingSystemCode, String(100)), nullable=True
    )
    rating: Mapped[int] = mapped_column("CRSE_RT_NBR", NUMBER(38))
    max_rating: Mapped[int] = mapped_column("CRSE_MAX_RT_NBR", NUMBER(38))
    comment: Mapped[Optional[str]] = mapped_column(
        "CRSE_RT_CMNT_TXT", String(4000), nullable=True
    )

    user: Mapped[OnlineLearningUser] = relationship(back_populates="course_ratings")
    course: Mapped[Course] = relationship(back_populates="ratings")


class Subject(Base):
    __tablename__ = "SBJCT"
    __repr_fields__ = ["id", "name", "slug", "source_system_id"]

    id: Mapped[str] = mapped_column("SBJCT_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_subject_id: Mapped[str] = mapped_column(
        "SRC_SYS_SBJCT_CD", String(255)
    )
    html_description: Mapped[str] = mapped_column("SBJCT_HTML_DES", String(4000))
    name: Mapped[str] = mapped_column("SBJCT_NM", String(100))
    slug: Mapped[Optional[str]] = mapped_column(
        "SBJCT_SLUG_TXT", String(4000), nullable=True
    )
    image_url: Mapped[Optional[str]] = mapped_column(
        "SBJCT_IMAGE_URL_TXT", String(4000), nullable=True
    )
    um_subject_name: Mapped[str] = mapped_column("UM_SBJCT_NM", String(100))

    source_system: Mapped[SourceSystem] = relationship(back_populates="subjects")
    courses: Mapped[List["Course"]] = relationship(
        secondary=course_subject, back_populates="subjects"
    )
    bundles: Mapped[List["Bundle"]] = relationship(
        secondary=bundle_subject, back_populates="subjects"
    )


class Bundle(Base):
    __tablename__ = "BUNDLE"
    __repr_fields__ = ["id", "name", "type", "slug", "source_system_id"]

    id: Mapped[str] = mapped_column("BUNDLE_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_bundle_id: Mapped[str] = mapped_column(
        "SRC_SYS_BUNDLE_CD", String(255)
    )
    source_system_bundle_instance_id: Mapped[str] = mapped_column(
        "SRC_SYS_BUNDLE_INSTC_CD", String(255)
    )
    type: Mapped[str] = mapped_column("BUNDLE_TYP_CD", String(50))
    name: Mapped[str] = mapped_column("BUNDLE_NM", String(500))
    description: Mapped[Optional[str]] = mapped_column(
        "BUNDLE_DES", String(4000), nullable=True
    )
    html_description: Mapped[Optional[str]] = mapped_column(
        "BUNDLE_HTML_DES", String(4000), nullable=True
    )
    slug: Mapped[Optional[str]] = mapped_column(
        "BUNDLE_SLUG_TXT", String(4000), nullable=True
    )
    version: Mapped[Optional[int]] = mapped_column(
        "BUNDLE_VER_NBR", NUMBER(38), nullable=True
    )

    source_system: Mapped[SourceSystem] = relationship(back_populates="bundles")
    courses: Mapped[List["Course"]] = relationship(
        secondary=course_bundle, back_populates="bundles"
    )
    subjects: Mapped[List["Subject"]] = relationship(
        secondary=bundle_subject, back_populates="bundles"
    )
    bundle_memberships: Mapped[List["BundleMembership"]] = relationship(
        back_populates="bundle"
    )

    # TODO: Define TypeCode enum for Bundle.type values


class CourseInstance(Base):
    __tablename__ = "CRSE_INSTC"
    __repr_fields__ = ["id", "course_id", "source_system_course_instance_id"]

    id: Mapped[str] = mapped_column("CRSE_INSTC_KEY", CHAR(36), primary_key=True)
    course_id: Mapped[str] = mapped_column(
        "CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY")
    )
    source_system_course_instance_id: Mapped[str] = mapped_column(
        "SRC_SYS_CRSE_INSTC_CD", String(50)
    )
    created_date: Mapped[Optional[datetime]] = mapped_column(
        "CRSE_INSTC_CRT_DTTM", UTCDateTime, nullable=True
    )

    classes: Mapped[List[Class]] = relationship(back_populates="course_instance")
    course: Mapped[Course] = relationship(back_populates="course_instances")
    modules: Mapped[List["Module"]] = relationship(back_populates="course_instance")
    submodules: Mapped[List["Submodule"]] = relationship(
        back_populates="course_instance"
    )
    items: Mapped[List["Item"]] = relationship(back_populates="course_instance")


class Module(Base):
    __tablename__ = "MODULE"
    __repr_fields__ = [
        "id",
        "course_instance_id",
        "class_id",
        "source_system_module_id",
    ]

    id: Mapped[str] = mapped_column("MODULE_KEY", CHAR(36), primary_key=True)
    course_instance_id: Mapped[str] = mapped_column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    class_id: Mapped[str] = mapped_column(
        "CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY")
    )
    source_system_module_id: Mapped[str] = mapped_column(
        "SRC_SYS_MODULE_CD", String(500)
    )
    order: Mapped[Optional[int]] = mapped_column(
        "MODULE_ORDER_NBR", NUMBER(38), nullable=True
    )
    name: Mapped[Optional[str]] = mapped_column(
        "MODULE_NM", String(4000), nullable=True
    )
    description: Mapped[Optional[str]] = mapped_column(
        "MODULE_DES", String(4000), nullable=True
    )
    start_date: Mapped[Optional[datetime]] = mapped_column(
        "MODULE_START_DTTM", UTCDateTime, nullable=True
    )
    staff_access: Mapped[Optional[int]] = mapped_column(
        "STAFF_ACCSS_IND", NUMBER(1), nullable=True
    )

    course_instance: Mapped[CourseInstance] = relationship(back_populates="modules")
    class_: Mapped[Class] = relationship(back_populates="modules")
    submodules: Mapped[List["Submodule"]] = relationship(back_populates="module")


class Submodule(Base):
    __tablename__ = "SUBMODULE"
    __repr_fields__ = [
        "id",
        "course_instance_id",
        "class_id",
        "source_system_submodule_id",
    ]

    id: Mapped[str] = mapped_column("SUBMODULE_KEY", CHAR(36), primary_key=True)
    course_instance_id: Mapped[str] = mapped_column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    class_id: Mapped[str] = mapped_column(
        "CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY")
    )
    source_system_submodule_id: Mapped[str] = mapped_column(
        "SRC_SYS_SUBMODULE_CD", String(500)
    )
    parent_submodule_id: Mapped[str] = mapped_column(
        "PRNT_SUBMODULE_KEY", CHAR(36), ForeignKey("SUBMODULE.SUBMODULE_KEY")
    )
    module_id: Mapped[str] = mapped_column(
        "MODULE_KEY", CHAR(36), ForeignKey("MODULE.MODULE_KEY")
    )
    order: Mapped[Optional[int]] = mapped_column(
        "SUBMODULE_ORDER_NBR", NUMBER(38), nullable=True
    )
    name: Mapped[Optional[str]] = mapped_column(
        "SUBMODULE_NM", String(4000), nullable=True
    )
    staff_access: Mapped[Optional[int]] = mapped_column(
        "STAFF_ACCSS_IND", NUMBER(1), nullable=True
    )

    course_instance: Mapped[CourseInstance] = relationship(back_populates="submodules")
    class_: Mapped[Class] = relationship(back_populates="submodules")
    module: Mapped[Module] = relationship(back_populates="submodules")
    parent: Mapped["Submodule"] = relationship(backref="children", remote_side=[id])
    items: Mapped[List["Item"]] = relationship(back_populates="submodule")


class Item(Base):
    __tablename__ = "ITEM"
    __repr_fields__ = ["id", "course_instance_id", "source_system_item_id"]

    id: Mapped[str] = mapped_column("ITEM_KEY", CHAR(36), primary_key=True)
    course_instance_id: Mapped[str] = mapped_column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    class_id: Mapped[str] = mapped_column(
        "CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY")
    )
    source_system_item_id: Mapped[str] = mapped_column("SRC_SYS_ITEM_CD", String(500))
    submodule_id: Mapped[str] = mapped_column(
        "SUBMODULE_KEY", CHAR(36), ForeignKey("SUBMODULE.SUBMODULE_KEY")
    )
    order: Mapped[Optional[int]] = mapped_column(
        "ITEM_ORDER_NBR", NUMBER(38), nullable=True
    )
    name: Mapped[Optional[str]] = mapped_column("ITEM_NM", String(255), nullable=True)
    type_description: Mapped[Optional[str]] = mapped_column(
        "ITEM_TYP_DES", String(255), nullable=True
    )
    type_category_name: Mapped[Optional[str]] = mapped_column(
        "ITEM_TYP_CTGRY_NM", String(255), nullable=True
    )
    is_graded: Mapped[Optional[int]] = mapped_column(
        "ITEM_TYP_GRDD_IND", NUMBER(1), nullable=True
    )
    url: Mapped[Optional[str]] = mapped_column(
        "ITEM_URL_TXT", String(4000), nullable=True
    )

    course_instance: Mapped[CourseInstance] = relationship(back_populates="items")
    class_: Mapped[Class] = relationship(back_populates="items")
    submodule: Mapped[Submodule] = relationship(back_populates="items")
    assessments: Mapped[List["Assessment"]] = relationship(
        secondary=assessment_item, back_populates="items"
    )


class Assessment(Base):
    __tablename__ = "ASSESSMENT"
    __repr_fields__ = ["id", "source_system_assessment_id", "category"]

    id: Mapped[str] = mapped_column("ASSESSMENT_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_assessment_id: Mapped[str] = mapped_column(
        "SRC_SYS_ASSESSMENT_CD", String(500)
    )
    category: Mapped[str] = mapped_column("ASSESSMENT_CTGRY_CD", String(25))
    type: Mapped[Optional[str]] = mapped_column(
        "ASSESSMENT_TYP_CD", String(50), nullable=True
    )
    pass_percentage: Mapped[Optional[float]] = mapped_column(
        "PASS_PCT", BINARY_FLOAT, nullable=True
    )
    updated_date: Mapped[Optional[datetime]] = mapped_column(
        "UPDT_DTTM", UTCDateTime, nullable=True
    )
    ttl: Mapped[Optional[str]] = mapped_column(
        "ASSESSMENT_TTL_TXT", String(500), nullable=True
    )
    max_submissions: Mapped[Optional[int]] = mapped_column(
        "MAX_SBMSSN_NBR", NUMBER(38), nullable=True
    )
    version: Mapped[Optional[int]] = mapped_column(
        "VERSION_NBR", NUMBER(38), nullable=True
    )
    max_score: Mapped[Optional[float]] = mapped_column(
        "MAX_SCR", BINARY_FLOAT, nullable=True
    )

    source_system: Mapped[SourceSystem] = relationship(back_populates="assessments")
    items: Mapped[List[Item]] = relationship(
        secondary=assessment_item, back_populates="assessments"
    )
    assessment_questions: Mapped[List["AssessmentQuestion"]] = relationship(
        back_populates="assessment"
    )
    peer_assignment: Mapped["PeerAssignment"] = relationship(
        uselist=False, back_populates="assessment"
    )
    programming_assignment: Mapped["ProgrammingAssignment"] = relationship(
        uselist=False, back_populates="assessment"
    )
    quiz: Mapped["Quiz"] = relationship(uselist=False, back_populates="assessment")


class PeerAssignment(Base):
    __tablename__ = "PEER_ASSGN"

    id: Mapped[str] = mapped_column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    required_review_count: Mapped[Optional[int]] = mapped_column(
        "RQRD_RVW_NBR", NUMBER(38), nullable=True
    )
    required_reviewer_for_score_count: Mapped[Optional[str]] = mapped_column(
        "RQRD_RVWR_FOR_SCR_NBR", NUMBER(38), nullable=True
    )
    required_wait_for_score: Mapped[Optional[int]] = mapped_column(
        "RQRD_WAIT_FOR_SCR_DUR_NBR", NUMBER(38), nullable=True
    )

    assessment: Mapped[Assessment] = relationship(back_populates="peer_assignment")


class ProgrammingAssignment(Base):
    __tablename__ = "PGMMNG_ASSGN"

    id: Mapped[str] = mapped_column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    submission_type: Mapped[Optional[str]] = mapped_column(
        "SBMSSN_TYP_CD", String(50), nullable=True
    )
    instructions: Mapped[Optional[str]] = mapped_column(
        "INSTRN_TXT", String(4000), nullable=True
    )
    submission_builder_schema_type: Mapped[Optional[str]] = mapped_column(
        "SBMSSN_BLDR_SCHEMA_TYP_CD", String(50), nullable=True
    )
    submission_builder_schema: Mapped[Optional[str]] = mapped_column(
        "SBMSSN_BLDR_SCHEMA_TXT", String(4000), nullable=True
    )

    assessment: Mapped[Assessment] = relationship(
        back_populates="programming_assignment"
    )


class Quiz(Base):
    __tablename__ = "QUIZ"

    id: Mapped[str] = mapped_column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    feedback_config: Mapped[Optional[str]] = mapped_column(
        "FDBCK_CONFIG_CD", String(50), nullable=True
    )

    assessment: Mapped[Assessment] = relationship(back_populates="quiz")


class Question(Base):
    __tablename__ = "QSTN"
    __repr_fields__ = ["id", "source_system_id", "type"]

    id: Mapped[str] = mapped_column("QSTN_KEY", CHAR(36), primary_key=True)
    source_system_id: Mapped[str] = mapped_column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_question_id: Mapped[str] = mapped_column(
        "SRC_SYS_QSTN_CD", String(500)
    )
    type: Mapped[Optional[str]] = mapped_column(
        "QSTN_TYP_CD", String(50), nullable=True
    )
    prompt: Mapped[Optional[str]] = mapped_column(
        "PROMPT_TXT", String(4000), nullable=True
    )
    is_shuffled: Mapped[Optional[int]] = mapped_column(
        "SHUFFLE_IND", NUMBER(1), nullable=True
    )
    allows_partial_scoring: Mapped[Optional[int]] = mapped_column(
        "PRTL_SCRNG_IND", NUMBER(1), nullable=True
    )
    default_incorrect_feedback: Mapped[Optional[str]] = mapped_column(
        "DFLT_INCORRECT_FDBCK_TXT", String(4000), nullable=True
    )
    feedback_text: Mapped[Optional[str]] = mapped_column(
        "FDBCK_TXT", String(4000), nullable=True
    )

    source_system: Mapped[SourceSystem] = relationship(back_populates="questions")
    assessment_questions: Mapped[List["AssessmentQuestion"]] = relationship(
        back_populates="question"
    )


class AssessmentQuestion(Base):
    __tablename__ = "ASSESSMENT_QSTN"
    __repr_fields__ = ["id", "source_system_id", "type"]

    assessment_id: Mapped[str] = mapped_column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    question_id: Mapped[str] = mapped_column(
        "QSTN_KEY", CHAR(36), ForeignKey("QSTN.QSTN_KEY"), primary_key=True
    )
    cuepoint: Mapped[Optional[int]] = mapped_column(
        "CUEPOINT_NBR", NUMBER(38), nullable=True
    )
    order: Mapped[Optional[int]] = mapped_column("ORDER_NBR", NUMBER(38), nullable=True)
    weight: Mapped[Optional[int]] = mapped_column("WT_NBR", NUMBER(38), nullable=True)
    max_attempts: Mapped[Optional[int]] = mapped_column(
        "MAX_ATTMPT_NBR", NUMBER(38), nullable=True
    )
    show_answer: Mapped[Optional[str]] = mapped_column(
        "SHOW_ANSWR_CD", String(100), nullable=True
    )
    has_group_access: Mapped[Optional[int]] = mapped_column(
        "GRP_ACCSS_IND", NUMBER(1), nullable=True
    )

    assessment: Mapped[Assessment] = relationship(back_populates="assessment_questions")
    question: Mapped[Question] = relationship(back_populates="assessment_questions")


class BundleMembership(Base):
    __tablename__ = "BUNDLE_MBRSHP"
    __repr_fields__ = ["id", "bundle_id", "user_id"]

    id: Mapped[str] = mapped_column("BUNDLE_MBRSHP_KEY", CHAR(36), primary_key=True)
    bundle_id: Mapped[str] = mapped_column(
        "BUNDLE_KEY", CHAR(36), ForeignKey("BUNDLE.BUNDLE_KEY")
    )
    user_id: Mapped[str] = mapped_column(
        "USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY")
    )
    role_code: Mapped[str] = mapped_column(
        "MBRSHP_ROLE_CD", EnumType(RoleCode, String(50))
    )
    begin_date: Mapped[datetime] = mapped_column("MBRSHP_BEGIN_DTTM", UTCDateTime)
    eligible_certificate_type: Mapped[Optional[str]] = mapped_column(
        "ELIG_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    awarded_certificate_type: Mapped[Optional[str]] = mapped_column(
        "AWARD_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    cert_award_date: Mapped[Optional[datetime]] = mapped_column(
        "CERT_AWARD_DTTM", UTCDateTime, nullable=True
    )
    payment_source_code: Mapped[Optional[str]] = mapped_column(
        "PMT_SRC_CD", EnumType(PaymentSourceCode, String(50)), nullable=True
    )

    bundle: Mapped[Bundle] = relationship(back_populates="bundle_memberships")
    user: Mapped[OnlineLearningUser] = relationship(back_populates="bundle_memberships")


class CourseMembership(Base):
    __tablename__ = "CRSE_MBRSHP"
    __repr_fields__ = ["id", "course_id", "user_id"]

    id: Mapped[str] = mapped_column("CRSE_MBRSHP_KEY", CHAR(36), primary_key=True)
    course_id: Mapped[str] = mapped_column(
        "CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY")
    )
    user_id: Mapped[str] = mapped_column(
        "USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY")
    )
    role_code: Mapped[str] = mapped_column(
        "MBRSHP_ROLE_CD", EnumType(RoleCode, String(50))
    )
    begin_date: Mapped[datetime] = mapped_column("MBRSHP_BEGIN_DTTM", UTCDateTime)
    eligible_certificate_type: Mapped[Optional[str]] = mapped_column(
        "ELIG_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    awarded_certificate_type: Mapped[Optional[str]] = mapped_column(
        "AWARD_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    cert_award_date: Mapped[Optional[datetime]] = mapped_column(
        "CERT_AWARD_DTTM", UTCDateTime, nullable=True
    )
    payment_source_code: Mapped[Optional[str]] = mapped_column(
        "PMT_SRC_CD", EnumType(PaymentSourceCode, String(50)), nullable=True
    )
    overall_grade: Mapped[Optional[int]] = mapped_column(
        "GRD_PCT", NUMBER(12, 9), nullable=True
    )

    course: Mapped[Course] = relationship(back_populates="course_memberships")
    user: Mapped[OnlineLearningUser] = relationship(back_populates="course_memberships")
